# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Plugin installer plugins call... to install themselves."""

import os
import shutil
import sys

from gettext import gettext as _

from dob_bright.config.app_dirs import AppDirs, get_appdirs_subdir_file_path

from .clickux.plugin_group import PLUGINS_DIRNAME

__all__ = (
    'install_plugin',
)


# ***
# *** Install-Plugin helper.
# ***

def install_plugin(package_module_path, package_plugin_name):
    """
    Helper function for Dob plugins to use to install themselves.

    For an example, see:

        https://github.com/hotoffthehamster/dob-plugin-example

    and search PyPI for matching "dob-plugin-\*" projects.
    """
    def _install_plugin():
        src_plugin = must_src_plugin_path()
        dst_plugin = must_dst_target_path()
        symlink_or_copy_plugin_or_die(src_plugin, dst_plugin)

    def must_src_plugin_path():
        src_plugin = os.path.join(
            os.path.dirname(package_module_path),
            '..',
            PLUGINS_DIRNAME,
            package_plugin_name,
        )
        assert os.path.exists(src_plugin)
        return src_plugin

    def must_dst_target_path():
        dst_plugin = get_appdirs_subdir_file_path(
            file_basename=package_plugin_name,
            dir_dirname=PLUGINS_DIRNAME,
            appdirs_dir=AppDirs.user_config_dir,
        )
        if os.path.exists(dst_plugin):
            print(_("The plugin is already installed!"))
            # Not really an error to already be installed,
            # so return not nonzero.
            sys.exit(0)
        return dst_plugin

    def symlink_or_copy_plugin_or_die(src_plugin, dst_plugin):
        try:
            symlink_or_copy_plugin(src_plugin, dst_plugin)
        except Exception as err:
            exit_error_unknown(err)

    def symlink_or_copy_plugin(src_plugin, dst_plugin):
        try:
            symlink_plugin(src_plugin, dst_plugin)
        except NotImplementedError:
            on_windows_copy_plugin(src_plugin, dst_plugin)

    def symlink_plugin(src_plugin, dst_plugin):
        os.symlink(src_plugin, dst_plugin, target_is_directory=False)
        print(_(
            "Successfully installed plugin (using symlink) to: {}"
        ).format(dst_plugin))

    def on_windows_copy_plugin(src_plugin, dst_plugin):
        shutil.copyfile(src_plugin, dst_plugin)
        print(_(
            "Successfully copied plugin to: {}"
        ).format(dst_plugin))
        print(_(
            "NOTE: To upgrade the plugin after upgrading the package, "
            "run this same command again."
        ))

    def exit_error_unknown(err):
        print(_(
            "OOPS! Something bad happened: {}"
        ).format(str(err)))
        print(_(
            "Please report this bug to the plugin author! Thanks!!"
            " (And for what it's worth, sorry!)"
        ))
        sys.exit(1)

    _install_plugin()




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import os
import shutil
import sys

from gettext import gettext as _

from nark import __file__ as nark___file__
from nark.helpers.app_dirs import ensure_directory_exists

from dob_bright.help_newbs import NEWBIE_HELP_WELCOME
from dob_bright.termio import (
    attr,
    click_echo,
    dob_in_user_exit,
    dob_in_user_warning,
    fg,
    highlight_value
)

from . import __arg0name__

__all__ = (
    'control',
    'downgrade',
    'upgrade',
    'version',
    'latest_version',
    'upgrade_legacy_database_file',
    'upgrade_legacy_database_instructions',
    # Private:
    #  '_barf_legacy_database',
    #  '_instruct_upgrade',
    #  '_upgrade_legacy_database_file',
)


def control(controller):
    err_msg = None
    success = False
    versioned = version(controller, silent_check=True, must=False)
    if not versioned:
        # MAYBE/2018-05-18: (lb): Also check if legacy database!
        success = controller.store.migrations.control()
        if success is None:
            err_msg = _('Something went wrong! Sorry!!')
    if not success and not err_msg:
        err_msg = _('The database is already versioned!')
    if err_msg:
        dob_in_user_exit(err_msg)


def downgrade(controller):
    version(controller, silent_check=True, must=True)
    response = controller.store.migrations.downgrade()
    assert response is not None
    if not response:
        dob_in_user_exit(_('The database is already at the earliest version'))


def upgrade(controller):
    version(controller, silent_check=True, must=True)
    response = controller.store.migrations.upgrade()
    assert response is not None
    if not response:
        dob_in_user_exit(_('The database is already at the latest version'))


def version(controller, silent_check=False, must=True):
    db_version = controller.store.migrations.version()
    if db_version is not None:
        if not silent_check:
            click_echo(db_version)
        return db_version
    elif must:
        _barf_legacy_database(controller)
    return None


def latest_version(controller, silent_check=False, must=True):
    latest_ver = controller.store.migrations.latest_version()
    if latest_ver is not None:
        if not silent_check:
            click_echo(latest_ver)
        return latest_ver
    elif must:
        _barf_legacy_database(controller)
    return None


# ***

UPGRADE_INSTRUCTIONS = _(
    """
If your database was created long ago by the original hamster,
`hamster-applet`, run the legacy database upgrade script, e.g.,

  {mintgreen}{prog_name} store upgrade-legacy \\
    {legacy_path}{reset}

Be sure to make a backup first, in case something goes wrong!
"""
)
# 2020-03-31: I had a link to the project page here but there's
# no additional information there, so why bother. It read,
#
#   See the project page for more information:
#
#       https://github.com/hotoffthehamster/nark
#
# As for actual upgrade help online, there's only one spot, but
# it says nothing more than what we've said here. It's at
#
#   https://dob.readthedocs.io/en/latest/usage.html#upgrade-hamster

# USER: If your database was created more recently by the new `hamster-lib`
# rewrite (2016-2017), there are 3 database changes to make. (lb): I did not
# add an upgrade path from hamster-lib because it seemed like that project
# was never quite released. So I assume no one's using it?


def upgrade_legacy_database_instructions(controller):
    """"""
    nark_path = os.path.dirname(os.path.dirname(nark___file__))
    up_legacy_rel = os.path.join(
        'migrations', 'legacy', 'upgrade_hamster-applet_db.sh',
    )
    up_legacy_path = os.path.join(nark_path, up_legacy_rel)
    db_path = controller.config['db.path']
    instructions = UPGRADE_INSTRUCTIONS.format(
        prog_name=__arg0name__,
        legacy_path="~/.local/share/hamster-applet/hamster.db",
        db_path=db_path,
        nark_path=nark_path,
        up_legacy_path=up_legacy_path,
        # FIXME: (lb): Replace hardcoded styles. Assign from styles.conf. #styling
        mintgreen=(fg('spring_green_2a') + attr('bold')),
        reset=attr('reset'),
    )
    return instructions


def _barf_legacy_database(controller):
    """"""
    prefix = '''
If this is a legacy database, upgrade and register the database.
'''.strip()
    msg1 = prefix + upgrade_legacy_database_instructions(controller)
    msg2 = _('The database is not versioned!')
    msg2 = '{}{}{}'.format(fg('red_3b'), msg2, attr('reset'))
    msg = '{}\n{}'.format(msg2, msg1)
    click_echo(msg)
    sys.exit(1)


# ***

def upgrade_legacy_database_file(ctx, controller, file_in, force):
    """"""
    db_path = controller.config['db.path']

    def _upgrade_legacy_database_file():
        if file_in is None:
            echo_help()
            return None
        else:
            copy_and_upgrade()
            return True

    def copy_and_upgrade():
        copy_into_place()
        run_upgrade()
        run_migrations()
        click_echo(
            _('Seeded data store at {}')
            .format(highlight_value(db_path))
        )

    def copy_into_place():
        if os.path.exists(db_path) and os.path.samefile(db_path, file_in.name):
            return
        controller.must_unlink_db_path(force=force)
        must_copy_new_db_to_db_path()

    def must_copy_new_db_to_db_path():
        try:
            db_dir = os.path.dirname(db_path)
            ensure_directory_exists(db_dir)
            shutil.copyfile(file_in.name, db_path)
        except Exception as err:
            msg = _('Failed to copy new database to ‘{}’').format(str(err))
            dob_in_user_warning(msg)

    def echo_help():
        if controller.is_germinated:
            click_echo(ctx.get_help())
        else:
            _instruct_upgrade(ctx, controller)

    def run_upgrade():
        try:
            controller.store.migrations.legacy_upgrade_from_hamster_applet(db_path)
        except Exception as err:
            dob_in_user_exit(_('Upgrade failed! ERROR: {}').format(str(err)))

    def run_migrations():
        db_version = controller.store.migrations.version()
        assert db_version is None
        control(controller)
        db_version = controller.store.migrations.version()
        assert db_version == 0
        response = True
        while response:
            response = controller.store.migrations.upgrade()

    return _upgrade_legacy_database_file()


def _instruct_upgrade(ctx, controller):
    click_echo(
        '\n{}\n{}'.format(
            NEWBIE_HELP_WELCOME(ctx),
            upgrade_legacy_database_instructions(controller),
        )
    )




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import os
import random
import re
import sys

from gettext import gettext as _

import click_hotoffthehamster as click
from click_hotoffthehamster.parser import split_arg_string
from nark.helpers.parsing import (
    ParserMissingActivityException,
    ParserMissingDatetimeOneException,
    ParserMissingDatetimeTwoException
)
from nark.items.fact import Fact

__all__ = ('tab_complete', )


TIME_HINT_MAP = {
    # SYNC_ME: RE_TIME_HINT, TIME_HINT_MAP, and @generate_add_fact_command's.
    'on': 'verify_none',
    'now': 'verify_none',  # 'on' alias
    'from': 'verify_both',
    'between': 'verify_both',  # 'from' alias
    'at': 'verify_start',
    'to': 'verify_end',
    'until': 'verify_end',  # 'to' alias
    'then': 'verify_then',
    'still': 'verify_still',
    'after': 'verify_none',
    'next': 'verify_none',  # 'after' alias
}


# FIXME/2018-05-17 18:04: On `dob import <tab>`, showing commands, not files! why??

def tab_complete(controller):
    def _complete_cmd(controller):
        args, incomplete = _do_complete()
        if len(args) < 1:
            return

        try:
            time_hint = TIME_HINT_MAP[args[0]]
        except KeyError:
            # Not an add_fact command! Fall-back to Click's complete.
            return

        for item in _get_choices(args[1:], incomplete, time_hint):
            click.echo(item)

    # NOTE: (lb): This fcn. was copied from click/_bashcomplete.py::do_complete().
    #       Then it was modified.
    def _do_complete():
        try:
            cwords = split_arg_string(os.environ['COMP_WORDS'])
        except KeyError:
            controller.client_logger.error(_(
                'Expected environ "COMP_WORDS" not set.'
            ))
            sys.exit(1)
        try:
            cword = int(os.environ['COMP_CWORD'])
        except ValueError:
            # This should only happen if your testing the complete command.
            assert os.environ['COMP_CWORD'] == ''
            cword = 0
        args = cwords[1:cword]
        try:
            incomplete = cwords[cword]
        except IndexError:
            incomplete = ''
        return (args, incomplete)

    def _get_choices(args, incomplete, time_hint):
        choices = []
        try:
            # Meh. Probably don't need FactDressed.
            fact, _err = Fact.create_from_factoid(args, time_hint=time_hint)
        except ParserMissingDatetimeOneException as err:
            choices = _choices_datetimes(controller, incomplete, time_hint, err)
        except ParserMissingDatetimeTwoException as err:
            choices = _choices_datetimes(controller, incomplete, time_hint, err)
        except ParserMissingActivityException:
            # See if we should suggest tags, activities, or activity-categories.
            if (not incomplete) or (not re.match(r'''^['"]?[#@]''', incomplete)):
                choices = choices_activities(
                    controller, incomplete, whitespace_ok=False,
                )
            else:
                choices = choices_tags(
                    controller, incomplete, whitespace_ok=False,
                )
        else:
            choices = [
                'Now_You_Write_A_Description',
                'Where-do-hamsters-go-after-graduation?__Hamsterdam!',
            ]

        return choices

    # The shim to the inline main(), above.
    _complete_cmd(controller)


def _choices_datetimes(controller, incomplete, time_hint, parser_err):
    """Suggest times."""
    now = controller.now
    # Show friendly usage reminders.
    # - For 'verify_start' (dob-at), show now and very recent times.
    # - For 'verify_end' (dob-to-/-until), show now (and very recent).
    # - For 'verify_both', with show less freshly recent for start,
    #   and show more recent times for second, end time.
    # NOTE: We underscore because Bash complete splits words,
    #       and thankfully the friendly `dateparser` understand this
    #       (for the most part; e.g., 'three years ago' works, but
    #       not 'three_years_ago', though '3_years_ago' works, go figure).
    friendly_at_hints = [
        '10_minutes_ago',
        '5_seconds_ago',
        'at_1_pm',
        'an_hour_ago',
        'noon',
        'midnight',
    ]
    friendly_from_hints = [
        # (lb): The shell splits completes on whitespace, so use underscores.
        # FIXME: (lb): Update parser to convert underscores back to whitespace
        #        before passing to human friendly time parser.
        'yesterday_at_3_PM',
        '1_week_ago',
        '3_years_ago',
        'Monday_at_16:44',
        'at_noon'
        '1_week_ago_at_midnight',
        # Hahaha, futuristic!
        # (lb): This is an invalid option. But it's funny!
        'tomorrow',
    ]

    # Show one friendly date example, randomly selected.
    # (lb): Using random.shuffle, rather than random.choice. Can't remember why.
    random.shuffle(friendly_at_hints)
    random.shuffle(friendly_from_hints)
    if time_hint == 'verify_start':
        friendly_hint = friendly_at_hints[:2]
    elif time_hint == 'verify_end':
        friendly_hint = ['now']
    elif time_hint == 'verify_both':
        if isinstance(parser_err, ParserMissingDatetimeOneException):
            friendly_hint = friendly_from_hints[:2]
        else:
            assert isinstance(parser_err, ParserMissingDatetimeTwoException)
            friendly_hint = friendly_at_hints[:2]
            friendly_hint.append('now')

    choices = [
        now.strftime('%Y-%m-%d'),
        now.strftime('%H:%M'),
        # Bash will split the completion word in a space,
        # so use \S+ form of YYYY-MM-DD and %H:M.
        # MEH: dateparser recognizes almost any delimiter
        # between the YYYY-MM-DD and the %H:%M, except
        # '-' and '+' indicate time zone. 'T' seems fine.
        now.strftime('%Y-%m-%dT%H:%M'),
    ]
    choices += friendly_hint
    if incomplete:
        choices = [
            ch for ch in choices if not incomplete or ch.startswith(incomplete)
        ]
    return choices


def choices_tags(controller, incomplete='', whitespace_ok=False):
    """Suggest tags."""
    # Grab the last 20 or so tags used (by Facts, chronologically),
    # and also the most used (top ten) 10 or so tags used by all Facts.
    #
    # FIXME: Make these limits settable (via config?).
    # FIXME: Can we cycle through the various sort options?
    #        E.g., if user TABs, show list of previous used tags.
    #          And if user TABs a second time, show list of most
    #          use tags. TAB again, another sort option, etc.
    # MAGIC_NUMBERS: 21 and 13, eh. Arbitrary limits.
    tags_counts = controller.tags.get_all_by_usage(sort_cols=('start',), limit=21)
    tags_counts += controller.tags.get_all_by_usage(sort_cols=('usage',), limit=13)

    choices = [
        '@{}'.format(tag.name) for tag, _uses, _span
        in tags_counts if not incomplete or tag.name.startswith(incomplete[1:])
    ]
    # MEH: We cull, even though we set limit above, so total count might be even smaller.
    #   (We could solve at the SQL query level, but who wants to go to the trouble?)
    if not whitespace_ok:
        choices = [tag for tag in choices if ' ' not in tag]
    return choices


def choices_activities(
    controller,
    incomplete='',
    whitespace_ok=False,
):
    """Suggest activities."""

    acty = controller.activities.get_all_by_usage(sort_cols=('start'))

    choices = [
        '{}@{}'.format(
            act.name, act.category.name if act.category else ''
        ) for act, count in acty
    ]

    # (lb): Bash complete doesn't handle spaces well, so ignore
    # those activities@categories with and spaces.
    if not whitespace_ok:
        choices = [atc for atc in choices if ' ' not in atc]
    if incomplete:
        choices = [atc for atc in choices if atc.startswith(incomplete)]

    # FIXME/2018-05-15 23:35: Caller should specify max, or just deal with
    # all results. E.g., tab-complete only wants so many. E.g., --ask only
    # wants as many can fit screen (or all, if I implement pagination). (lb)
    max_choices = 50
    if len(choices) > max_choices:
        choices = choices[:50]

    return choices




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""``dob usage`` commands."""

from gettext import gettext as _

from pedantic_timedelta import PedanticTimedelta

from dob_bright.reports.render_results import render_results

__all__ = ('generate_usage_table', )


def generate_usage_table(
    controller,
    results,
    name_header=None,
    name_fmttr=lambda item: item.name if item else '<NULL>',
    show_usage=False,
    show_duration=False,
    output_format='table',
    table_type='texttable',
    max_width=-1,
    output_path=None,
):
    def generate_usage_table():
        staged = []
        max_width_tm_value = 0
        max_width_tm_units = 0
        for item, count, duration in results:
            (
                tm_fmttd, tm_scale, tm_units,
            ) = PedanticTimedelta(days=duration or 0).time_format_scaled()
            value, units = tm_fmttd.split(' ')
            max_width_tm_value = max(max_width_tm_value, len(value))
            max_width_tm_units = max(max_width_tm_units, len(units))
            staged.append((item, count, tm_fmttd))

        rows = []
        for item, count, tm_fmttd in staged:
            value, units = tm_fmttd.split(' ')
            span = '{0:>{1}} {2:^{3}}'.format(
                value, max_width_tm_value, units, max_width_tm_units,
            )

            rows.append((name_fmttr(item), count, span))

        culled = cull_results(rows)

        headers = prepare_headers()

        render_results(
            controller,
            results=culled,
            headers=headers,
            output_format=output_format,
            table_type=table_type,
            max_width=max_width,
            output_path=output_path,
        )

    def cull_results(results):
        if show_usage and show_duration:
            return results

        resulted = []
        for result in results:
            newr = [result[0]]
            if show_usage:
                newr.append(result[1])
            if show_duration:
                newr.append(result[2])
            resulted.append(newr)
        return resulted

    def prepare_headers():
        first_header = name_header
        if first_header is None:
            first_header = _("Name")
        headers = [first_header]
        if show_usage:
            headers.append(_("Uses"))
        if show_duration:
            headers.append(_("Total Time"))
        return headers

    generate_usage_table()




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from dob_bright.termio import dob_in_user_exit

__all__ = (
    'error_exit_no_results',
)


# ***

def error_exit_no_results(item_type, more_msg=None):
    # MAYBE: (lb): Should we exit on error?
    #              Print message but not error?
    #              Print empty table?
    #              Only print if no facts whatsoever,
    #                as opposed to no facts matching query?
    #        I'd say either print message on stderr and exit, or
    #        print empty table on stdout, so that a tool in the
    #        pipeline would only see empty string or empty table
    #        on stdin, and never a message, so that post-processing
    #        won't die on unexpected error message string.
    more_msg = ': {}'.format(more_msg) if more_msg else ''
    msg = _('No {} were found for the specified query{}.').format(item_type, more_msg)
    # (lb): I sorta like the message, not table, on no results,
    # because you can tell quicker that nothing was found. I.e.,
    # when I see an empty tell, I spend a nanosecond scanning it
    # for rows, because when I see table headers, I expect to see
    # table rows! So I'm kinda liking printing a message, not table.
    dob_in_user_exit(msg)




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from functools import update_wrapper

__all__ = (
    'post_processor',
)


# ***

def post_processor(func):
    """
    """

    def wrapper(ctx, controller, *args, **kwargs):
        # Ensure that plugins are loaded. Plugins may have functions
        # decorated with @Controller.post_processor to be called.
        # - ctx.parent is a click_hotoffthehamster.core.Context, and
        #   ctx.parent.command is <ClickAliasableBunchyPluginGroup run>.
        ctx.parent.command.ensure_plugged_in(controller)
        facts = func(ctx, controller, *args, **kwargs)
        controller.post_process(controller, facts, show_plugin_error=None)

    return update_wrapper(wrapper, func)




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Defines the main Click Group."""

import sys
from functools import update_wrapper

import click_hotoffthehamster as click

# BREADCRUMB: PROFILING
from nark.helpers.dev.profiling import profile_elapsed, timefunct

from dob_bright.termio import click_echo, echo_exit
from dob_bright.termio.paging import ClickEchoPager

from .clickux import help_strings
from .clickux.aliasable_bunchy_plugin import ClickAliasableBunchyPluginGroup
from .controller import DobController
from .copyright import echo_copyright

__all__ = (
    'pass_controller',
    'pass_controller_context',
    'dob_versions',
    'run',
    # Private:
    #  'CONTEXT_SETTINGS',
)


# Profiling: Controller is made during Command.invoke via click.MultiCommand.invoke.
# Profiling: Controller calls _get_store: ~ 0.173 secs.
pass_controller = click.make_pass_decorator(DobController, ensure=True)


# ***

def pass_controller_context(func):
    @pass_controller
    @click.pass_context
    def _pass_context(ctx, controller, *args, **kwargs):
        controller.ctx = ctx
        func(ctx, controller, *args, **kwargs)

    return update_wrapper(_pass_context, func)


# ***
# *** [VERSION] Version command helper.
# ***

def dob_versions(include_all=False):
    '''Return CLI version information, either for this package, or all HOTH packages.
    '''
    vers = ''
    include_head = include_all
    import importlib
    # MAYBE/2020-04-01: Add config_decorator and pedantic_timedelta.
    hothlibs = ['dob']
    if include_all:
        hothlibs += [
            'dob_viewer',
            'dob_prompt',
            'dob_bright',
            'nark',
        ]
    minlen = max([len(name) for name in hothlibs])
    for hothlib in hothlibs:
        mod = importlib.import_module(hothlib, package=None)
        vers += '\n' if vers else ''
        vers += '{name:{minlen}s} version {vers}'.format(
            minlen=minlen,
            name=mod.__package_name__,
            vers=mod.get_version(include_head=include_head),
        )
    return vers


# ***
# *** [CLICK ROOT CONTEXT] Twiddle default Context behavior.
# ***

CONTEXT_SETTINGS = dict(
    # Tell Click to plumb the -h and --help options.
    help_option_names=['-h', '--help'],
    # But also tell Click not to invoke the help callback,
    # which it would otherwise do immediately when it sees
    # the help option.
    #
    # By default, Click processes help immediately when spotted.
    #
    # This has a few consequences:
    #
    # - If --help is seen before the subcommand in args, the run() method
    #   below is not called, because Click handles the --help before invoking
    #   the callback, and the Click help callback prints the help and exits.
    #   As such, color and paging will not be setup properly for root help.
    #   However, for subcommand help, our run() will be called, and then color
    #   and paging will be setup properly for the subcommand help!
    #
    # - If the user wants help on a subcommand, Click forces them to specify
    #   the --help option after the command, i.e., `dob --help command` would
    #   normally print the same thing as `dob --help`. But we'd rather print
    #   the help for the subcommand specified. (This also makes such obvious
    #   commands like `dob help command` work as expected as well!)
    #
    # So tell Click not to process help upon sight, but to mark a flag.
    # Later, we'll look for help_option_spotted on the root context.
    help_option_fallthrough=True,
    # The --help width, used for one-line summaries, as well as the longer
    # reformatted help documentation.
    # MAGIC_NUMBER: Set reasonably wide, but the help strings themselves
    #               will likely limit their widths to 89 or so.
    max_content_width=101,
)


# ***
# *** [BASE COMMAND GROUP] One Group to rule them all.
# ***

# (lb): Use invoke_without_command so `dob -v` works, otherwise Click's
# Group (MultiCommand ancestor) does not allow it ('Missing command.').
@click.group(
    # (lb): Use our Plugin group class , which dynamically loads pluggable
    # commands from user's, e.g., ~/.config/dob/plugins directory. That class
    # derives from an Aliasable group class, which empowers us to assign command
    # name aliases. That class, in turn, derives from Click's base Group class.
    cls=ClickAliasableBunchyPluginGroup,
    invoke_without_command=True,
    help=help_strings.RUN_HELP_OVERVIEW,
    context_settings=CONTEXT_SETTINGS,
)
# (lb): Also include version to avoid RuntimeError, ha!
@click.version_option(message=dob_versions(), version='')
# (lb): Hide -v: version_option adds help for --version, so don't repeat ourselves.
@click.option('-v', is_flag=True, help=help_strings.VERSION_HELP, hidden=True)
# (lb): Note that universal --options must com before the sub command.
# FIXME: Need universal options in cmd_options? Or can I apply to Groups?
#          These aren't recognized by other fcns...
#        OH! These have to come *before* the command??
@click.option('-V', '--verbose', is_flag=True,
              help=help_strings.GLOBAL_OPT_VERBOSE)
@click.option('-VV', '--verboser', is_flag=True, hidden=True,
              help=help_strings.GLOBAL_OPT_VERBOSER)
@click.option('-X', '--color/--no-color', default=None,
              help=help_strings.GLOBAL_OPT_COLOR_NO_COLOR)
@click.option('-P', '--pager/--no-pager', default=None,
              help=help_strings.GLOBAL_OPT_PAGER_NO_PAGER)
@click.option('-C', '--config', multiple=True, metavar='KEY=VALUE',
              help=help_strings.GLOBAL_OPT_CONFIG)
# (lb): We could use `type=click.File('r')` here. Or not.
@click.option('-F', '--configfile', metavar='PATH',
              help=help_strings.GLOBAL_OPT_CONFIGFILE)
# Profiling: pass_controller appears to take ~ ¼ seconds.
@timefunct('run: create Controller [_get_store]')
@pass_controller
@click.pass_context
# NOTE: @click.group transforms this func. definition into a callback that
#       we use as a decorator for the top-level commands (see: @run.command).
def run(ctx, controller, v, verbose, verboser, color, pager, config, configfile):
    """General context run right before any of the commands."""

    def _run(ctx, controller, show_version):
        """
        Do stuff before running the command.

        Setup paging, if paging.
        Enable/disable color, per config.
        No longer show a banner.
        Show version and exit, if user specified -v option.
        Setup up loggers.
        """
        profile_elapsed('To dob:    run')
        controller.ensure_config(ctx, configfile, *config)
        _setup_tty_options(ctx, controller)
        _run_handle_banner()
        _run_handle_version(ctx, show_version)
        _run_handle_without_command(ctx)
        controller.setup_logging(verbose, verboser)

    def _setup_tty_options(ctx, controller):
        # If piping output, Disable color and paging.
        # MAYBE: (lb): What about allowing color for outputting to ANSI file? Meh.
        use_color = color
        if use_color is None and not sys.stdout.isatty():
            controller.config['term.use_pager'] = False
            controller.config['term.use_color'] = False
        _setup_tty_paging(controller)
        _setup_tty_color(ctx, controller)
        controller.pre_apply_style_conf()

    def _setup_tty_paging(controller):
        use_pager = pager
        if use_pager is None:
            # None if --pager nor --no-pager specified,
            # so fallback to what's in the user config.
            use_pager = controller.config['term.use_pager']
        ClickEchoPager.set_paging(use_pager)

    def _setup_tty_color(ctx, controller):
        use_color = color
        controller.setup_tty_color(use_color)
        # We'll set the Click Context object's color attribute, too, but note
        # that only Click's echo() method strips color based on that attribute;
        # the format_help() methods do not scrub ANSI codes, so we still need
        # to be color-aware ourselves. (So setting ctx.color here is more of a
        # formality than something that actually does anything.)
        ctx.color = color

    def _run_handle_banner():
        # (lb): I find the greeting annoying, and somewhat boastful.
        #   It's not that I'm against self-promotion, per se -- and I
        #   recognize that the GPL instructs open source software to
        #   always output a license, on every invocation -- but I like
        #   the clean aesthetics of not showing it. Though I suppose
        #   it's easy enough just to make a config option for it....
        if not controller.config['term.show_greeting']:
            return
        echo_copyright()
        click_echo()

    def _run_handle_version(ctx, show_version):
        if show_version:
            echo_exit(ctx, dob_versions(include_all=False))

    def _run_handle_without_command(ctx):
        # Because we set invoke_without_command, we have to check ourselves
        # if invoked without any command, in which case show the help.
        # Note this code originally, naively checked `if len(sys.argv) == 1`,
        # but the context (which is always root when run() is invoked) knows
        # if there's a subcommand specified, so check that attribute instead.
        if ctx.invoked_subcommand is None:
            echo_exit(ctx, ctx.get_help())

    # Shim to the private run() functions.

    _run(ctx, controller, show_version=v)




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2019-2020 Landon Bouma. All rights reserved.
#
# This program is free software:  you can redistribute it  and/or  modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any later version  (GPLv3+).
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY;  without even the implied warranty of MERCHANTABILITY or  FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU  General  Public  License  for  more  details.
#
# If you lost the GNU General Public License that ships with this software
# repository (read the 'LICENSE' file), see <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from click_hotoffthehamster.exceptions import MissingParameter

from config_decorator.key_chained_val import KeyChainedValue

# Load all the upstream config packages to ensure `dob config dump`, etc.,
# show all config values.
from nark import config as nark_config  # noqa: F401 '<>' imported but unused
from dob_bright import config as dob_bright_config  # noqa: F401
from dob_viewer import config as dob_viewer_config  # noqa: F401

from dob_bright.crud.interrogate import run_editor_safe
from dob_bright.termio import click_echo, dob_in_user_exit
from dob_bright.termio.config_table import echo_config_decorator_table

__all__ = (
    'echo_config_table',
    'echo_config_value',
    'edit_config_file',
    'write_config_value',
    # PRIVATE:
    #  'echo_config_value_setting',
    #  'echo_config_value_section',
    #  'exit_error_no_setting',
    #  'fetch_config_object',
    #  'fetch_config_objects',
    #  'error_exit_not_one',
    #  'config_parts_pop_value',
    #  'must_parts',
    #  'must_be_config_setting',
)


# *** [DUMP] TABLE

def echo_config_table(
    controller,
    section,
    keyname,
    output_format,
    table_type,
):
    """"""
    parts = list(filter(None, (section, keyname)))
    conf_objs = fetch_config_objects(controller, parts)
    include_hidden = section and keyname
    echo_config_decorator_table(
        controller,
        conf_objs,
        output_format=output_format,
        table_type=table_type,
        include_hidden=include_hidden,
    )


# *** [EDIT] CONFIG

def edit_config_file(controller):
    run_editor_safe(filename=controller.configurable.config_path)


# *** [GET] ECHO

def echo_config_value(ctx, controller, parts):
    must_parts(ctx, parts, _('“KEYNAME”'))
    section_or_setting = fetch_config_object(controller, parts)
    if isinstance(section_or_setting, KeyChainedValue):
        echo_config_value_setting(section_or_setting)
    else:
        echo_config_value_section(section_or_setting)


def echo_config_value_setting(setting):
    click_echo(setting.value)


def echo_config_value_section(section):
    counts = []
    if section._sections:
        counts.append('{} {}'.format(len(section._sections), _('sections')))
    if section._key_vals or not section._sections:
        counts.append('{} {}'.format(len(section._key_vals), _('settings')))
    dob_in_user_exit(
        _('Configuration section “{}” contains {}.').format(
            section._name,
            _('and').join(counts),
        )
    )


# *** [SET] CONFIG

def write_config_value(ctx, controller, parts):
    parts, value = config_parts_pop_value(ctx, parts)
    section_or_setting = fetch_config_object(controller, parts)
    setting = must_be_config_setting(section_or_setting)
    try:
        setting.value = value
    except ValueError as err:
        dob_in_user_exit(str(err))
    controller.write_config(skip_unset=True)


# *** [GET/SET] FETCH

def fetch_config_object(controller, parts):
    conf_objs = fetch_config_objects(controller, parts)
    if len(conf_objs) == 1:
        return conf_objs[0]
    error_exit_not_one(parts, conf_objs)


def fetch_config_objects(controller, parts):
    try:
        conf_objs = controller.configurable.find_all(parts)
    except AttributeError:
        exit_error_no_setting(parts)
    except KeyError:
        # Raised by configurable on unknown part(s).
        exit_error_no_setting(parts)
    return conf_objs


def error_exit_not_one(parts, conf_objs):
    dotted = '.'.join(parts)
    if len(conf_objs) > 1:
        dob_in_user_exit(
            _('ERROR: Too many config objects named: “{}”').format(dotted)
        )
    else:
        # FIXME/2019-11-17 02:18: Should errors be paged if --pager?
        # - See also/Consolidate: barf_and_exit, echo_exit, dob_in_user_exit.
        dob_in_user_exit(
            _('ERROR: Not a config object: “{}”.').format(dotted)
        )


# *** [SECTION] [KEYNAME] [VALUE] VALIDATING

def config_parts_pop_value(ctx, parts):
    parts = list(parts)
    must_parts(ctx, parts, _('“KEYNAME” “VALUE”'), param_type='arguments')
    value = parts.pop()
    must_parts(ctx, parts, _('“VALUE”'))
    return parts, value


def must_parts(ctx, parts, param_hint, param_type='argument'):
    parts = list(parts)
    if not parts:
        raise MissingParameter(
            ctx=ctx, param_hint=param_hint, param_type=param_type,
        )
    return parts


def must_be_config_setting(section_or_setting):
    if not isinstance(section_or_setting, KeyChainedValue):
        dob_in_user_exit(_(
            'ERROR: Not a configuration setting: “{}”.'.format(section_or_setting._name)
        ))
    return section_or_setting


def exit_error_no_setting(parts):
    dob_in_user_exit(_(
        'ERROR: Not a configuration setting: “{}”.'.format('.'.join(parts))
    ))




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from functools import update_wrapper

from dob_bright.crud.fact_dressed import FactDressed
from dob_bright.termio import click_echo, dob_in_user_exit, echo_block_header

from .. import __arg0name__, migrate

__all__ = (
    'induct_newbies',
    'insist_germinated',
    # Private:
    #  'backend_integrity',
)


def backend_integrity(func):
    """
    Verify that data in the database is integrit.

    (lb): I wonder if the backend should enforce this better.
    In any case, telling the user at the CLI level is better
    than not telling the user at all, I suppose.
    """

    def wrapper(ctx, controller, *args, **kwargs):
        version_must_be_latest(controller)
        time_must_be_gapless(controller)
        func(ctx, controller, *args, **kwargs)

    # ***

    def version_must_be_latest(controller):
        db_version = migrate.version(
            controller, silent_check=True, must=True,
        )
        latest_version = migrate.latest_version(
            controller, silent_check=True, must=True,
        )
        if db_version != latest_version:
            assert db_version < latest_version
            msg = _(
                'Expected database to be same version as latest migration.'
                ' {} != {}'
                '\nTrying running `{} migrate up`'
            ).format(db_version, latest_version, __arg0name__)
            dob_in_user_exit(msg)

    # ***

    def time_must_be_gapless(controller):
        endless_facts = controller.facts.endless()
        barf_on_endless_facts(endless_facts)

    def barf_on_endless_facts(endless_facts):
        # There can be only 1. Or none.
        if not endless_facts or len(endless_facts) == 1:
            return

        for fact in endless_facts:
            # FIXME/2018-05-18: (lb): Make this prettier.
            echo_block_header(_('Endless Fact Found!'))
            click_echo()
            click_echo(fact.friendly_diff(fact))
            click_echo()
        msg = _(
            'Found saved fact(s) without start time and/or end time.'
            '\nSee list of offending Facts above.'
            # MAYBE/2018-05-23 17:05: (lb): We could offer an easy way out, e.g.,
            #   '\n\nTry, e.g.,\n\n  {} edit {} --end now'.format(__arg0name__, ...)
        )
        dob_in_user_exit(msg)

    # ***

    return wrapper


def insist_germinated(func):
    """
    """

    def wrapper(ctx, controller, *args, **kwargs):
        controller.insist_germinated(fact_cls=FactDressed)
        func(ctx, controller, *args, **kwargs)

    return wrapper


def induct_newbies(func):
    """
    """
    @insist_germinated
    @backend_integrity
    def wrapper(*args, **kwargs):
        func(*args, **kwargs)

    return update_wrapper(wrapper, func)




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from dob_bright.termio import ascii_table

from dob import cmds_list


class TestActivities(object):
    """Unit tests for the ``activities`` command."""

    def test_tabulate_list_activities_no_category(
        self, controller, activity, mocker, capsys,
    ):
        """Make sure command works if activities do not have a category associated."""
        activity.category = None
        mocker.patch.object(
            controller.activities, 'get_all', return_value=[activity],
        )
        mocker.patch.object(
            ascii_table.tabulate,
            'tabulate',
            return_value='{}, {}'.format(activity.name, None),
        )
        cmds_list.activity.list_activities(
            controller,
            output_format='table',
            # Specify any one of 'tabulate' types.
            table_type='rst',
        )
        out, err = capsys.readouterr()
        assert out.startswith(activity.name)
        assert ascii_table.tabulate.tabulate.call_args[0] == ([(activity.name, None)],)

    def test_list_activities_no_category(self, controller, activity, mocker, capsys):
        """Make sure command works if activities do not have a category associated."""
        activity.category = None
        mocker.patch.object(
            controller.activities, 'get_all', return_value=[activity],
        )
        mocker.patch.object(ascii_table.texttable.Texttable, 'add_rows')
        mocker.patch.object(
            ascii_table.texttable.Texttable,
            'draw',
            return_value='{}, {}'.format(activity.name, None),
        )
        cmds_list.activity.list_activities(
            controller,
            output_format='table',
            table_type='texttable',
        )
        out, err = capsys.readouterr()
        assert out.startswith(activity.name)
        add_rows_call_args = ascii_table.texttable.Texttable.add_rows.call_args
        # The first row is the headers. Second row is activity, category passed.
        assert add_rows_call_args[0][0][1] == (activity.name, None)
        draw_call_args = ascii_table.texttable.Texttable.draw.call_args
        assert draw_call_args[0] == ()

    def test_list_activities_no_filter(
        self, controller, activity, mocker, capsys,
    ):
        """Make sure activity name and category are displayed if present."""
        mocker.patch.object(
            controller.activities, 'get_all', return_value=[activity],
        )
        cmds_list.activity.list_activities(controller)
        out, err = capsys.readouterr()
        assert activity.name in out
        assert activity.category.name in out

    def test_list_activities_using_category(
        self, controller, activity, mocker, capsys,
    ):
        """Make sure the search term is passed on."""
        mocker.patch.object(
            controller.activities, 'query_process_results', return_value=[activity],
        )
        mocker.patch.object(
            controller.activities,
            'query_filter_by_category_name',
            return_value=activity.category.name,
        )
        cmds_list.activity.list_activities(
            controller,
            match_categories=[activity.category],
        )
        out, err = capsys.readouterr()
        assert controller.activities.query_filter_by_category_name.called
        controller.activities.query_filter_by_category_name.assert_called_with(
            activity.category,
        )
        assert activity.name in out
        assert activity.category.name in out

    def test_list_activities_with_search_term(
        self, controller, activity, mocker, capsys,
    ):
        """Make sure the search term is passed on."""
        mocker.patch.object(
            controller.activities, 'gather', return_value=[activity],
        )
        cmds_list.activity.list_activities(
            controller,
            # Defaults.
            output_format='table',
            show_usage=True,
            show_duration=True,
            # The one we're testing.
            search_terms=[activity.category.name],
        )
        out, err = capsys.readouterr()
        assert controller.activities.gather.called
        # Mock's assert_called_with checks arguments exactly, but we
        # just care to check that the search_term was there.
        query_terms = controller.activities.gather.call_args[0][0]
        assert query_terms.search_terms == [activity.category.name]
        # The category is a pass-through list_activities **kwarg, so
        # won't be in the kwargs.
        assert query_terms.match_categories == []
        assert activity.name in out
        assert activity.category.name in out




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob

"""
Packaging instruction for setup tools.

Refs:

  https://setuptools.readthedocs.io/

  https://packaging.python.org/en/latest/distributing.html

  https://github.com/pypa/sampleproject
"""

from setuptools import find_packages, setup

# *** Package requirements.

requirements = [
    # "textwrap, but savvy to ANSI colors"
    #  https://github.com/jonathaneunice/ansiwrap
    'ansiwrap >= 0.8.4, < 1',
    # Vocabulary word pluralizer.
    #  https://github.com/ixmatus/inflector
    'Inflector >= 3.0.1, < 4',
    # https://github.com/mnmelo/lazy_import
    'lazy_import >= 0.2.2, < 1',
    # Tabulate is one of the many table formatter choices.
    #  https://bitbucket.org/astanin/python-tabulate
    'tabulate >= 0.8.7, < 1',
    # Texttable is one of the many table formatter choices.
    #  https://github.com/bufordtaylor/python-texttable
    'texttable >= 1.6.2, < 2',

    # *** HOTH packages.

    # (lb): Click may be the best optparser of any language I've used.
    #  https://github.com/pallets/click
    #    'click',
    #  - Still, had to make one adjustment, and too impatient to ask for a pull...
    #  https://github.com/hotoffthehamster/click
    'click-hotoffthehamster >= 7.1.1, <= 7.1.2',
    # Indispensable aliases support for Click.
    #  Upstream at: https://github.com/click-contrib/click-aliases
    #  Released at: https://github.com/hotoffthehamster/click-hotoffthehamster-alias
    'click-hotoffthehamster-alias == 1.0.2',
    # Elapsed timedelta formatter, e.g., "1.25 days".
    # - Imports as `pedantic_timedelta`.
    #  https://github.com/hotoffthehamster/human-friendly_pedantic-timedelta
    'human-friendly_pedantic-timedelta >= 2.0.0, <= 2.0.10',
    # The heart of Hamster. (Ye olde `hamster-lib`).
    #  https://github.com/hotoffthehamster/nark
    'nark > 3.2.1, < 3.2.3',  # I.e., release 3.2.2, or whatever dev's running.
    # The controller, config, and common output and error tossing code.
    #  https://github.com/hotoffthehamster/dob-bright
    'dob-bright > 1.2.2, < 1.2.4',  # I.e., release 1.2.3, or dev's intermediate.
    # The so-called Facts "carousel" chrono-viewer.
    #  https://github.com/hotoffthehamster/dob-viewer
    'dob-viewer > 1.2.2, < 1.2.4',  # I.e., release 1.2.3, or dev's intermediate.
]

# *** Minimal setup() function -- Prefer using config where possible.

# (lb): Most settings are in setup.cfg, except identifying packages.
# (We could find-packages from within setup.cfg, but it's convoluted.)

setup(
    # Run-time dependencies installed on `pip install`. To learn more
    # about "install_requires" vs pip's requirements files, see:
    #   https://packaging.python.org/en/latest/requirements.html
    install_requires=requirements,

    # Specify which package(s) to install.
    # - Without any rules, find_packages returns, e.g.,
    #     ['dob', 'tests', 'tests.dob']
    # - With the 'exclude*' rule, this call is essentially:
    #     packages=['dob']
    packages=find_packages(exclude=['tests*']),

    # Tell setuptools to determine the version
    # from the latest SCM (git) version tag.
    #
    # Note that if the latest commit is not tagged with a version,
    # or if your working tree or index is dirty, then the version
    # from git will be appended with the commit hash that has the
    # version tag, as well as some sort of 'distance' identifier.
    # E.g., if a project has a '3.0.0a21' version tag but it's not
    # on HEAD, or if the tree or index is dirty, the version might
    # be:
    #   $ python setup.py --version
    #   3.0.0a22.dev3+g6f93d8c.d20190221
    # But if you clean up your working directory and move the tag
    # to the latest commit, you'll get the plain version, e.g.,
    #   $ python setup.py --version
    #   3.0.0a31
    # Ref:
    #   https://github.com/pypa/setuptools_scm
    setup_requires=['setuptools_scm'],
    use_scm_version=True,
)




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from dob_bright.termio import dob_in_user_warning

__all__ = (
    'compile_and_eval_source',
)


def compile_and_eval_source(py_path):
    """"""
    def _compile_and_eval_source(py_path):
        with open(py_path, 'r') as py_text:
            eval_globals = compile_and_eval_module(py_text, py_path)
            return eval_globals

    def compile_and_eval_module(py_text, py_path):
        code = source_compile(py_text, py_path)
        if code is None:
            return {}
        eval_globals = globals()
        if eval_source_code(code, eval_globals, py_path):
            return eval_globals
        return {}

    def source_compile(py_text, py_path):
        try:
            code = compile(py_text.read(), py_path, 'exec')
        except Exception as err:
            code = None
            msg = _(
                'ERROR: Could not compile source file at "{}": {}'
            ).format(py_path, str(err))
            dob_in_user_warning(msg)
        return code

    def eval_source_code(code, eval_globals, py_path):
        # Pass py_path to code being eval'd, so it can orientate.
        # (lb): I tried passing `locals()` for second argument, but
        # then plugins complain `name 'ConfigRoot' is not defined`.
        eval_globals['__file__'] = py_path
        try:
            eval(code, eval_globals, eval_globals)
        except Exception as err:
            msg = _(
                'ERROR: Could not eval compiled source at "{}": {}'
            ).format(py_path, str(err))
            dob_in_user_warning(msg)
            return False
        else:
            return True

    return _compile_and_eval_source(py_path)




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from dob_bright.reports.render_results import render_results

from ..clickux.query_assist import error_exit_no_results

__all__ = ('list_tags', )


def list_tags(
    controller,
    output_format='table',
    table_type='texttable',
    max_width=-1,
    output_path=None,
    # These two --hide flags are ignored but specified to keep out of kwargs.
    show_usage=False,
    show_duration=False,
    **kwargs
):
    """
    List tag names, optionally filtered and formatted specially.

    Writes to stdout, or to the file specified by ``output_path``.
    """
    err_context = _('tags')

    results = controller.tags.get_all(**kwargs)

    results or error_exit_no_results(err_context)

    headers = (_("Tag Name"),)
    tag_names = []
    for tag in results:
        tag_names.append((tag.name,))

    render_results(
        controller,
        results=tag_names,
        headers=headers,
        output_format=output_format,
        table_type=table_type,
        max_width=max_width,
        output_path=output_path,
    )




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Click Group wrapper combines all our custom classes."""

from gettext import gettext as _

import click_hotoffthehamster as click

from .help_header import help_header_format

__all__ = (
    'ClickBetterHeadersGroup'
)


class ClickBetterHeadersGroup(click.Group):

    def __init__(self, *args, **kwargs):
        super(ClickBetterHeadersGroup, self).__init__(*args, **kwargs)

    @property
    def help_header_options(self):
        # (lb): Without having the caller pass us the ctx, so we can
        # check ctx.parent is None to know if this is the root command
        # or not, we can use our business knowledge to check the command
        # name, which is 'run', the name of the root callback in run_cli.
        # - I'd rather use ctx.parent is None, and not do a string compare,
        #   but this is dirt cheap and easy squeeze y.
        if self.name == 'run':
            return help_header_format(_('Global Options'))
        return help_header_format(_('Command Options'))

    @property
    def help_header_commands(self):
        # click.core's behavior: return 'Commands:' (with a colon;
        # but we prefer an underline instead).
        return help_header_format(_('Commands'))




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma,  2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from dob.cmds_list import fact as list_fact


class TestBasicRun(object):
    def test_basic_run(self, runner):
        """Make sure that invoking the command passes without exception."""
        result = runner()
        assert result.exit_code == 0


class TestSearchWithoutInitPrintsSetupMessage(object):
    def test_search_all(self, runner):
        """Running search command fails until dob conf and store are initialized."""
        result = runner(['search'])
        assert result.exit_code == 1
        assert "Let’s get you setup!" in result.stdout


class TestDobSearchCommandCallsFeatureHandler(object):
    def test_dob_search_command_calls(
        self,
        mocker,
        dob_runner,
        capsys,
    ):
        """Running search command fails until dob conf and store are initialized."""
        mocker.patch.object(list_fact, 'list_facts')
        result = dob_runner(['search'])
        assert result.exit_code == 0
        assert list_fact.list_facts.called
        # We stubbed list_facts, so there won't be any output.
        out, err = capsys.readouterr()
        assert not out and not err




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma, © 2015-2016 Eric Goller.  All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

import click_hotoffthehamster as click

__all__ = (
    'cmd_options_edit_item',
    'cmd_options_fact_add',
    'cmd_options_fact_dryable',
    'cmd_options_fact_edit',
    'cmd_options_fact_import',
    'cmd_options_factoid',
    'cmd_options_factoid_verify_none',
    'cmd_options_factoid_verify_start',
    'cmd_options_factoid_verify_end',
    'cmd_options_factoid_verify_both',
    'cmd_options_rule_name',
    'cmd_options_styles_internal',
    'cmd_options_styles_named',
)


# ***
# *** [ADD FACT/STOP FACT] Raw Factoid Option.
# ***

_option_factoid_text_parts = (
    '[<activity>[@<category>]]'
    ' [@<tag>...] [<description>]'
)


_cmd_options_factoid = [
    click.argument('factoid', nargs=-1, default=None,
                   metavar='[START_TIME] [to] [END_TIME] {}'.format(
                       _option_factoid_text_parts,
                   )),
]


def cmd_options_factoid(func):
    for option in reversed(_cmd_options_factoid):
        func = option(func)
    return func


_cmd_options_factoid_verify_none = [
    click.argument('factoid', nargs=-1, default=None,
                   metavar=_option_factoid_text_parts,
                   ),
]


def cmd_options_factoid_verify_none(func):
    for option in reversed(_cmd_options_factoid_verify_none):
        func = option(func)
    return func


_cmd_options_factoid_verify_start = [
    click.argument('factoid', nargs=-1, default=None,
                   metavar='[START_TIME] {}'.format(
                       _option_factoid_text_parts,
                   )),
]


def cmd_options_factoid_verify_start(func):
    for option in reversed(_cmd_options_factoid_verify_start):
        func = option(func)
    return func


_cmd_options_factoid_verify_end = [
    click.argument('factoid', nargs=-1, default=None,
                   metavar='[END_TIME] {}'.format(
                       _option_factoid_text_parts,
                   )),
]


def cmd_options_factoid_verify_end(func):
    for option in reversed(_cmd_options_factoid_verify_end):
        func = option(func)
    return func


_cmd_options_factoid_verify_both = [
    click.argument('factoid', nargs=-1, default=None,
                   metavar='START_TIME to END_TIME {}'.format(
                       _option_factoid_text_parts,
                   )),
]


def cmd_options_factoid_verify_both(func):
    for option in reversed(_cmd_options_factoid_verify_both):
        func = option(func)
    return func


# ***
# *** [ADD FACT] Options.
# ***

_cmd_options_fact_add_prefix = [
    click.option(
        '-e', '--editor', is_flag=True,
        help=_('Edit new Fact before saving, using Carousel, and Awesome Prompt.'),
    ),
]


_cmd_options_fact_add_and_edit = [
    click.option(
        '-d', '--edit-text', is_flag=True,
        help=_('Edit description using user’s preferred $EDITOR.'),
    ),
    click.option(
        '-a', '--edit-meta', is_flag=True,
        help=_('Ask for act@gory and tags using Awesome Prompt.'),
    ),
]


_cmd_options_fact_add_postfix = [
    # (lb): 2019-02-01: Current thinking is that conflicts are only okay
    # on add-fact, and only outside the context of the Carousel. So applies
    # to dob-add commands, but not to dob-import.
    click.option(
        '-y', '--yes', is_flag=True,
        help=_('Save conflicts automatically, otherwise ask for confirmation.'),
    ),
]


def cmd_options_fact_add(func):
    for option in reversed(
        _cmd_options_fact_add_prefix
        + _cmd_options_fact_add_and_edit
        + _cmd_options_fact_add_postfix
    ):
        func = option(func)
    return func


# ***
# *** [IMPORT FACTS] Options.
# ***

_cmd_options_fact_import = [
    # (lb): This is similar to dob-add's --edit, except the default is reversed.
    # - On dob-add, default is to not run Carousel; but on dob-import, it is.
    click.option(
        # Option skips carousel, opens Content in EDITOR, saves Fact on EDITOR exit.
        '-E', '--no-editor', is_flag=True,
        help=_('Skip interactive editor after import. Save Facts and exit.'),
    ),
]


def cmd_options_fact_import(func):
    for option in reversed(_cmd_options_fact_import):
        func = option(func)
    return func


# ***
# *** [EDIT FACT] Options.
# ***

_cmd_options_fact_no_editor_edit = [
    click.option(
        '-E', '--no-editor', is_flag=True,
        help=_('Skip interactive editor. Use $EDITOR and Awesome Prompt.'),
    ),
]


def cmd_options_fact_edit(func):
    for option in reversed(
        _cmd_options_fact_add_and_edit
        + _cmd_options_fact_no_editor_edit
    ):
        func = option(func)
    return func


# ***
# *** [ADD FACT/IMPORT FACT(S)] Shared Options.
# ***

_cmd_options_fact_dryable = [
    # (lb): The --dry option is not super useful if you have your store under
    # git control, because you can easily revert any changes; or if you setup a
    # test store under /tmp (i.e., `export XDG_DATA_HOME=/tmp/xxx/.local/share`).
    # It's really just more code to test!
    # MAYBE/2019-02-01: Remove the --dry option, and save a unittest?
    click.option(
        '--dry', is_flag=True,
        help=_('Dry run: do not make changes.'),
    ),
]


def cmd_options_fact_dryable(func):
    for option in reversed(_cmd_options_fact_dryable):
        func = option(func)
    return func


# ***
# *** [EDIT ITEM] Options.
# ***

_cmd_options_edit_item = [
    # User can indicate specific item to edit via its PK, otherwise default to latest.
    # FIXME/BACKLOG/2019-01-31: Could allow user to specify datetime instead of PK,
    #   e.g., `dob edit 2019-01-31` could bring up Fact at Noon on specific day (or
    #   midnight).
    click.argument('key', nargs=1, type=int, required=False),
    # (lb): User can specify specific Fact PK, a positive integer, or user
    # can specify an index relative to the last Fact, e.g., `dob edit -1`
    # (or even `dob edit -2`, though anything other than `dob edit -1` seems
    # useless, i.e., would a user ever really run `dob edit -5`?). In any case,
    # because the negative relative index starts with the dash '-' character,
    # Click will complain if it parses the argument as an option, e.g.,
    #   $ dob edit -1
    #   Error: no such option: -1
    # The user can double-dash to tell Click to stop option processing, e.g.,
    #   $ dob edit -- -1
    # The latter is somewhat clunky, so we can make '-1' an option.
    # Note that this doesn't solve the issue for -2, -3, etc., but really, who cares.
    click.option(
        '-1', 'latest_1', is_flag=True,
        help=_('Edit most recent Fact (latest complete, or active).'),
    ),
]


def cmd_options_edit_item(func):
    for option in reversed(_cmd_options_edit_item):
        func = option(func)
    return func


# ***
# *** [STYLES] Options.
# ***

_cmd_options_styles_internal = [
    click.option(
        '-i', '--internal', is_flag=True,
        help=_(
            'Print pristine or internal style settings, not raw styles.conf section. '
            'If the named style is also a section in styles.conf, '
            'this exclude comments, and it reorders the settings. '
            'Otherwise, if not in styles.conf, the matching internal style is located.'
        )
    ),
]


def cmd_options_styles_internal(func):
    for option in reversed(_cmd_options_styles_internal):
        func = option(func)
    return func


_cmd_options_styles_named = [
    click.argument('name', nargs=1, default='', metavar=_('[STYLE_NAME]')),
]


def cmd_options_styles_named(func):
    for option in reversed(_cmd_options_styles_named):
        func = option(func)
    return func


# ***
# *** [RULES] Options.
# ***

_cmd_options_rule_name = [
    click.argument('name', nargs=1, default='', metavar=_('[RULE_NAME]')),
]


def cmd_options_rule_name(func):
    for option in reversed(_cmd_options_rule_name):
        func = option(func)
    return func




# This file exists within 'dob':
#
#   https://github.com/hotoffthehamster/dob
#
# Copyright © 2018-2020 Landon Bouma. All rights reserved.
#
# 'dob' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'dob' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""dob-demo command"""

from gettext import gettext as _

import os
import tempfile
from datetime import timedelta
from functools import update_wrapper

from nark.items.activity import Activity
from nark.items.category import Category

from dob_bright.crud.fact_dressed import FactDressed

# We could use the wrapper method and tell it to Carousel:
#  from .facts.save_confirmed import prompt_and_save_confirmed
# Or we could just call the Carousel wrapper directly:
from dob_viewer.traverser.save_confirmer import prompt_and_save_confirmer

__all__ = (
    'demo_config',
    'demo_dob',
    # Private:
    # '_demo_prep'
)


def demo_config(func):
    """
    """
    def wrapper(controller, *args, **kwargs):
        tmpfile = _demo_prep(controller)
        func(controller, *args, **kwargs)
        # The temp. file is removed on exit regardless, but just to be clear:
        tmpfile.close()

    return update_wrapper(wrapper, func)


# ***

def demo_dob(controller):
    def _demo_dob():
        genator = DemoFactGenerator(controller)
        demo_facts = []
        # Save the facts to the temp. db., otherwise Carousel
        # won't quit on simple 'q', but will prompt user to save.
        for demo_fact in genator.demo_facts():
            saved_fact = controller.facts.save(demo_fact)
            demo_facts.append(saved_fact)

        prompt_and_save_confirmer(controller, edit_facts=demo_facts, dry=False)

        # Remove the temporary file that _demo_prep created, that we could
        # not let tempfile delete on garbage collection, because Windows.
        # Note that SQLAlchemy's db reference causes Windows to complain
        # on unlink, so explicitly close the session first. Otherwise, e.g.,:
        #   PermissionError: [WinError 32] The process cannot access the file
        #   because it is being used by another process:
        #     'C:\\Users\\IEUser\\AppData\\Local\\Temp\\dob-demo-vr5mi28q.sqlite'
        controller.store.session.close()
        os.unlink(controller.config['db.path'])

    return _demo_dob()


class DemoFactGenerator(object):
    """"""
    def __init__(self, controller):
        self.controller = controller
        self.create_actegories()
        self.populate_facts()

    # ***

    def create_actegories(self):
        self.create_categories()
        self.create_activities()

    def create_categories(self):
        self.cats = {
            'welcome': Category(name=_('Welcome')),
            'intermediate': Category(name=_('Intermediate Skills')),
        }

    def create_activities(self):
        self.acts = {
            'demo@intermediate': Activity(
                name=_('Demo'), category=self.cats['intermediate'],
            ),
            'demo@welcome': Activity(
                name=_('Demo'), category=self.cats['welcome'],
            ),
        }

    # ***

    def demo_facts(self):
        for demo_fact in self._demo_facts:
            yield demo_fact

    def populate_facts(self):
        self._demo_facts = []
        prev_fact = None
        for name in list(filter(lambda name: name.startswith('demo_fact_'), dir(self))):
            demo_fact = getattr(self, name)(prev_fact)
            self._demo_facts.append(demo_fact)
            prev_fact = demo_fact

    # ***

    def demo_fact_00(self, _prev_fact):
        return self.first_fact()

    def demo_fact_99(self, _prev_fact):
        return self.final_fact()

    def final_fact(self):
        demo_fact = FactDressed(
            start=self.controller.now - timedelta(hours=1),
            end=None,
            activity=self.acts['demo@welcome'],
            # tags=['dob-life', 'welcome', ],
            tags=['hello, dobber!', ],
            description=_('''
Welcome to the dob demo!

Do you want to learn the basics of dob? Then follow along!

* To quit at any time, press the "q" key.

Let’s get started! You’re looking at the last, final Fact.

* Press the "g" key twice ("gg") to go to the first Fact.
            '''.strip()),  # noqa: E501
        )
        return demo_fact

    def first_fact(self):
        since_time = self.controller.now - timedelta(days=14)
        until_time = since_time + timedelta(minutes=90)
        demo_fact = FactDressed(
            start=since_time,
            end=until_time,
            activity=self.acts['demo@welcome'],
            tags=['first-fact', ],
            description=_('''
Congratulations, you made it to the first Fact in the demo!

The "gg" command takes you to the first Fact in your database.

The "G" command (uppercase), similarly, takes you to the last Fact.

* Press "G" now to try it, then press "gg" to return here.

* To continue, press the "k" key to advance to the next Fact.
            '''.strip()),  # noqa: E501
        )
        return demo_fact

    # (lb): My apologies, this could be a maintenance pain, the way
    # I've named the methods, encoding their order therein. It's not
    # going to be easy to move sections around!

    def demo_fact_01(self, prev_fact):
        demo_fact = FactDressed(
            start=prev_fact.end,
            end=prev_fact.end + timedelta(minutes=66),
            activity=self.acts['demo@welcome'],
            tags=['learning fast', ],
            description=_('''
You're learning fast!

To go backward one Fact, press the "j" key.

* Press "j" now to try it, then press "k" to return here.

You can also use the left and right arrow keys to change Facts.

* To continue, press the right arrow key "→" to advance one Fact.
            '''.strip()),  # noqa: E501
        )
        return demo_fact

    def demo_fact_02(self, prev_fact):
        demo_fact = FactDressed(
            start=prev_fact.end,
            end=prev_fact.end + timedelta(hours=12),
            activity=self.acts['demo@intermediate'],
            tags=['choose your own demo #3', ],
            description=_('''
Wow! You can sure get around dob now!

You can also navigate dob by jumping Facts one day at a time.

* Press (uppercase) "K" to jump to the Fact 1 day from now.

  Return to this entry with (uppercase) "J".

* Then press (lowercase) "k" (or "→") to keep reading the demo.
            '''.strip()),  # noqa: E501
        )
        return demo_fact

    def demo_fact_03(self, prev_fact):
        demo_fact = FactDressed(
            start=prev_fact.end,
            end=prev_fact.end + timedelta(hours=9),
            activity=self.acts['demo@intermediate'],
            tags=['choose your own demo #5', ],
            # E501 line too long (✗✗✗ > 89 characters)
            description=_('''
You’ve become a pro at navigating between Facts!

Let’s learn how to navigate the *description*, which is what you’re reading.

You can move the cursor using "h" and "l", or using the up and down arrow keys.

* Press and hold "l" or the down arrow key "↓" until the cursor scrolls down the window.

  You should eventually see more text.




Hi! You made it!!

To move the cursor up, you can use either "h" or the up "↑" arrow key.

And you can use a three-button mouse wheel to scroll the description, too.

To scroll by the pageful, you can use Page Up and Page Down ("PgUp" and "PgDn").

* Press "PgUp" a few times and then "PgDn" back here to try for yourself.

To jump to the very beginning of the description, use "Home".

To jump to the very end, use "End".

* To keep reading the demo, press "End" now.








This is the end of the description!

* To keep learning, jump one day -- press "K" -- because you’ve read the next Fact.

  Hint: Look at the status bar below this text and you’ll see
  the "Fact ID #4" change to "Fact ID #6" after pressing "K".
            '''.strip()),  # noqa: E501
        )
        return demo_fact

    def demo_fact_04(self, prev_fact):
        demo_fact = FactDressed(
            start=prev_fact.end,
            end=prev_fact.end + timedelta(hours=6),
            activity=self.acts['demo@intermediate'],
            tags=['choose your own demo #4', ],
            description=_('''
Did you just press "K" to jump forward one day? Congrats!

Bonus Feature: If you press and hold "K" (don't do it now), dob will start jumping by weeks.

To continue the demo, jump back one day, to the last entry you read.

* Press "J" to jump back one day to continue the demo.
            '''.strip()),  # noqa: E501
        )
        return demo_fact

    def demo_fact_05(self, prev_fact):
        demo_fact = FactDressed(
            start=prev_fact.end,
            end=prev_fact.end + timedelta(hours=18),
            activity=self.acts['demo@intermediate'],
            tags=['choose your own demo #6', ],
            description=_('''
Before we can learn how to create and edit Facts, there are a few more basics!

(And remember to scroll down, using the down arrow key "↓", or "PgDn".)

Immediate Help
--------------

You can access the help at any time by pressing the question mark "?" key.

* Press "?" to view the online help now.

  Then, press "?" twice more to return here:

  - The second "?" will show you the second page of the help.

  - The third "?" will exit the help. (You can also press "q" to close the help.)

Gaps
----

* Press "→" or "l" to learn about Gap Facts.
            '''.strip()),  # noqa: E501
        )
        return demo_fact

    def demo_fact_06(self, prev_fact):
        demo_fact = FactDressed(
            start=prev_fact.end,
            end=prev_fact.end + timedelta(hours=18),
            activity=self.acts['demo@intermediate'],
            tags=['choose your own demo #7', ],
            description=_('''
Gap Facts
---------

Each Fact has a *start time* and an *end time*.

* You can see the start and end for this Fact in the headers shown above.

In dob, Facts are ordered chronologically.

* When you press "j", dob shows you the previous Fact in time.

  When you press "k", dob shows you the next Fact in time.

If there is a gap in time between two Facts, dob will add what's called a "Gap Fact".

A Gap Fact is simply a placeholder -- it is not part of your data... yet.

If you want, you can edit and save a Gap Fact, and then it will be stored.

But, if not, you can ignore the Gap Fact, and dob will not save it.

Example Gap
-----------

To see a Gap Fact for real, the next Fact in the demo starts 17 minutes later, after the end of this Fact.

* To continue the demo, move 2 Facts forward:

  - Press "k" (or "→") twice, but pause between presses to examine the Gap Fact.

    - The next Fact is a Gap Fact, and it is not technically part of the demo data.

    - You’ll notice that the Gap Fact has a special background color, sorta pink.

      You will also see that the status message, below this text, says “Gap Fact”.
            '''.strip()),  # noqa: E501
        )
        return demo_fact

    def demo_fact_07(self, prev_fact):
        since_time = prev_fact.end + timedelta(minutes=17)
        demo_fact = FactDressed(
            start=since_time,
            end=since_time + timedelta(hours=6),
            activity=self.acts['demo@intermediate'],
            tags=['After the Gap Fact #8', ],
            description=_('''
Gap Review
----------

As stated in the last step of the demo, the previous Fact has a sorta pinkish background because it's a *Gap Fact*. It is considered a temporary placeholder unless you edit and save it.

Speaking of editing, let’s talk about that finally!

* Press "k" (or "→") to learn how to edit Fact times.
            '''.strip()),  # noqa: E501
        )
        return demo_fact

    def demo_fact_08(self, prev_fact):
        demo_fact = FactDressed(
            start=prev_fact.end,
            end=prev_fact.end + timedelta(hours=4.75),
            activity=self.acts['demo@intermediate'],
            tags=['choose your own demo #9', ],
            description=_('''
Edit Time
---------

To edit the start time or end time of the current Fact, move the cursor to the corresponding widget.

Tab Focus
---------

To move the cursor between the description, the start time, and the end time, press "Tab".

* Press "Tab" three times now.

  You’ll move the cursor to the start time, to the end time, and then back here.

You can also press "Shift-Tab" to move the cursor in the reverse direction.

Send Focus
----------

Instead of "Tab" or "Shift-Tab":

* You can press "s" to send the cursor to the start time widget.

* You can press "e" to send the cursor to the end time widget.

* From within the start or end time widget, you can press "d" to send the cursor to the description, or content area.

Manual Time
-----------

When the cursor is in the start or end time widget, you can edit the value as expected.

* Use the left and right arrow keys to move the cursor back and forth across the time value.

* Use the "Backspace" key to delete characters, and use the number keys and punctuation to enter a new time.

* Press "Enter" (or "Tab" away) to apply the new time.

Audit Time
----------

Note that changing the start or end time can affect surrounding Facts!

* If you change the start time to an earlier time, dob may edit the Fact before, and might change its end time to match the start time.

  Likewise, for an edited end time, dob may edit the start time of the following Fact.

* To protect the integrity of your data, dob will not let you specify a time that completely shadows any other Facts' time.

  For instance, you can not change the start time of a Fact so that it precedes the start of any other Fact.

  * If you try, dob will correct your mistake and set the edited time to the closest time allowed to not obliterate any Facts.

Quick Time
----------

You can quickly edit time using "Shift", "Ctrl", and "Ctrl-Shift" modifiers with the left and right arrow keys.

* Press "Shift-left" to decrement the start time one minute.

* Press "Shift-right" to increment the start time one minute.

* Press "Ctrl-left" to decrement the end time one minute.

* Press "Ctrl-right" to increment the end time one minute.

* Press "Ctrl-Shift-left" to decrement both the start time and the end time by one minute.

* Press "Ctrl-Shift-right" to increment both the start time and the end time by one minute.

You can also press-and-hold to continuously increment or decrement by one minute intervals -- which grow longer than one minute as you keep holding.

Fact Editing
------------

To read about editing the Description, Activity & Category, and Tags, move forward 1 Fact.

* Press "k" (or "→") to continue the demo.
            '''.strip()),  # noqa: E501
        )
        return demo_fact

    def demo_fact_09(self, prev_fact):
        demo_fact = FactDressed(
            start=prev_fact.end,
            end=prev_fact.end + timedelta(hours=4.75),
            activity=self.acts['demo@intermediate'],
            tags=['choose your own demo #10', ],
            description=_('''
Meta Edits
----------

Besides editing the start and end time, you can edit the
Fact Description, the Activity & Category, and the Tags.

* To edit the Fact description using your preferred editor, press "d".

  You may need to set the EDITOR environment variable before running dob, e.g.,

    $ export EDITOR=/usr/bin/vim
    $ dob demo

  Make and save any changes, then exit the editor to return to dob.

* To edit the Activity & Category, press "a".

  dob will show you an advanced, interactive interface.

  You can type a new ``activity@category``.

  You can also choose from names you’ve used on other Facts, and you can
  sort the drop down lists in various ways to make it easier to choose
  Activity and Category names.

* To edit Tags, press "t".

  dob will show you an advanced tag editor.

  You can enter new tags, or you can choose from existing tags.

  And you can sort the list of recommended tags in different ways,
  including by recently used, or by most used, etc., using the F-keys.

  Read the help at the bottom of the tags display for more.

Save
----

* Press "k" (or "→") to learn how to save edits.
            '''.strip()),  # noqa: E501
        )
        return demo_fact

    def demo_fact_10(self, prev_fact):
        demo_fact = FactDressed(
            start=prev_fact.end,
            end=prev_fact.end + timedelta(hours=4.75),
            activity=self.acts['demo@intermediate'],
            tags=['choose your own demo #11', ],
            description=_('''
Saving
------

dob does not automatically save your edits.

* To save changes you’ve made, press "Ctrl-s".

If you have any unsaved edits, the "q" shortcut to quit will not work.

* You can force-quit instead using "Ctrl-Q" (uppercase), and dob will prompt you to exit without saving.

You can also press "u" repeatedly to undo all of your changes (and then "q" will work again).

Undo/Redo
---------

* Press "k" (or "→") to continue the demo and learn how to undo and redo.
            '''.strip()),  # noqa: E501
        )
        return demo_fact

    def demo_fact_11(self, prev_fact):
        demo_fact = FactDressed(
            start=prev_fact.end,
            end=prev_fact.end + timedelta(hours=4.75),
            activity=self.acts['demo@intermediate'],
            tags=['choose your own demo #12', ],
            description=_('''
Undo/Redo
---------

dob supports a typical undo and redo workflow.

* Press "Ctrl-z" (or "u") to undo the last change.

  Press again to keep undoing.

* Press "Ctrl-y" (or "r", or "Ctrl-r") to redo the last undo.

  Press again to keep redoing.

  (Note that the multitude of aliases is because parity with Vim.)

If there are no more undos or redos, dob will display a message in the status bar.

Ctrl-c/-v
---------

The next Fact explains how copy and paste work in dob.

* Press "k" (or "→") to continue the demo.
            '''.strip()),  # noqa: E501
        )
        return demo_fact

    def demo_fact_12(self, prev_fact):
        demo_fact = FactDressed(
            start=prev_fact.end,
            end=prev_fact.end + timedelta(hours=4.75),
            activity=self.acts['demo@intermediate'],
            tags=['choose your own demo #13', ],
            description=_('''
Copy/Paste
----------

* Press "Ctrl-c" to copy a Fact.

  Then navigate to another Fact and press "Ctrl-v" to paste it.

* Press "Ctrl-x" instead to cut a Fact.

  The original Fact (that you copied) will be removed when you paste it (and dob will show you an empty Gap Fact in the display where the original Fact used to be).

* If you want to copy and paste a smaller part of the Fact, it's easy:

  * Press "A" followed by "Ctrl-c" to copy the Activity & Category.

  * Press "T" followed by "Ctrl-c" to copy the current Fact's tags.

  * Press "D" followed by "Ctrl-c" to copy the current Fact's description.

  Then paste like normal, using "Ctrl-v".

Split/Merge
-----------

* Press "k" (or "→") to continue the demo.
            '''.strip()),  # noqa: E501
        )
        return demo_fact

    def demo_fact_13(self, prev_fact):
        demo_fact = FactDressed(
            start=prev_fact.end,
            end=prev_fact.end + timedelta(hours=4.75),
            activity=self.acts['demo@intermediate'],
            tags=['choose your own demo #14', ],
            description=_('''
Split/Merge/Delete
------------------

* Press "Alt-p" to split the current Fact into two Facts.

  Then edit each Fact separately, including adjusting the times accordingly.

* Press "Alt-e" to delete the current Fact.

  dob will remove the Fact and replace it in the display with a Gap Fact.

  When you save, the removed Fact will be marked deleted in the data store.

* Press "Alt-m" followed by "→" to merge the current Fact with the next Fact.

  Press "Alt-m" followed by "←" to merge the current Fact with the prior Fact.

Start using dob
---------------

To get up and running, read the next Fact.

* Press "k" ("→") to learn how to dob yourself.
            '''.strip()),  # noqa: E501
        )
        return demo_fact

    def demo_fact_14(self, prev_fact):
        demo_fact = FactDressed(
            start=prev_fact.end,
            end=prev_fact.end + timedelta(hours=4.75),
            activity=self.acts['demo@intermediate'],
            tags=['choose your own demo #15', ],
            description=_('''
Getting Started
---------------

The easiest way to start dobbin' is from scratch.

Run the init command to create a config and a database:

  $ dob init

And then fire up the Carousel (whatever this interface is called that you're using):

  $ dob edit

If you'd like to upgrade from an existing Hamster database, that's easy, too:

  $ dob init
  $ legacy_db=~/.local/share/hamster-applet/hamster.db
  $ dob store upgrade-legacy ${legacy_db}
  $ dob edit

You can always run this demo, even after setting up your own database.

You're a pro!
-------------

* Press "k" ("→") to conclude the demo.
            '''.strip()),  # noqa: E501
        )
        return demo_fact

    def demo_fact_15(self, prev_fact):
        demo_fact = FactDressed(
            start=prev_fact.end,
            end=prev_fact.end + timedelta(hours=4.75),
            activity=self.acts['demo@intermediate'],
            tags=['choose your own demo #16', 'end of the ride', ],
            description=_('''
You made it to the end of the demo!! -- But there's much more you can do with dob:

  - Import Facts from raw text (collected while, say, on vacation).
  - Generate reports (e.g., show time spent on each Activity).
  - Find and install (and write your own) plugins.
  - Export data to other APIs (e.g., upload time to JIRA at your job).

This demo gives you the gist of dob. We hope you enjoyed it!!

Happy dobbin!
            '''.strip()),  # noqa: E501
        )
        return demo_fact


# ***

def _demo_prep(controller):
    """"""

    def __demo_prep():
        tmpfile = create_temporary_file()
        demoize_config(db_path=tmpfile.name)
        controller.standup_store(fact_cls=FactDressed)
        return tmpfile

    def create_temporary_file():
        # (lb): Windows does not allow NamedTemporaryFile to be opened twice,
        # so close here, without allowing it to be deleted, and then the store
        # will reopen it. We'll remove it when the user ends the demo.
        tmpfile = tempfile.NamedTemporaryFile(
            prefix='dob-demo-',
            suffix='.sqlite',
            delete=False,
        )
        tmpfile.close()
        return tmpfile

    def demoize_config(db_path):
        controller.config['db.orm'] = 'sqlalchemy'
        controller.config['db.engine'] = 'sqlite'
        controller.config['db.path'] = db_path
        # For completeness, reset the others.
        controller.config['db.host'] = ''
        controller.config['db.port'] = ''
        controller.config['db.name'] = ''
        controller.config['db.user'] = ''

    return __demo_prep()

